{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Day 19](https://adventofcode.com/2019/day/19) puzzle input is an IntCode program (mine is [here](puzzle_input/day19.txt)) which will models a sort of beam, and when provided with an x/y co-ordinate will output whether the beam is in effect at that position. Part 1 involves finding the number of positions for x and y in the range 0-49 which are affected by the beam. Part 2 involves figuring out the first 100x100 section which can be wholly contained within the beam (position of the top-left corner)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "opcode_add = 1\n",
    "opcode_mul = 2\n",
    "opcode_read = 3\n",
    "opcode_write = 4\n",
    "opcode_jump_true = 5\n",
    "opcode_jump_false = 6\n",
    "opcode_lt = 7\n",
    "opcode_eq = 8\n",
    "opcode_rebase = 9\n",
    "\n",
    "opcode_terminate = 99\n",
    "\n",
    "mode_position = 0\n",
    "mode_immediate = 1\n",
    "mode_relative = 2\n",
    "\n",
    "class IntCodeCpu:\n",
    "    def __init__(self, memory_image):\n",
    "        self.memory = [ x for x in memory_image ] # copy memory image, in case it's reused\n",
    "        self.stalled = True\n",
    "        self.input_buffer = None\n",
    "        self.output_buffer = None\n",
    "        self.pc = 0\n",
    "        self.initialise_opcodes()\n",
    "        self.offset = 0\n",
    "        self.done = False\n",
    "        \n",
    "    def start(self, input_buffer, output_buffer, noun=None, verb=None):\n",
    "        self.input_buffer = input_buffer\n",
    "        self.output_buffer = output_buffer\n",
    "\n",
    "        if noun:\n",
    "            self.memory[1] = noun\n",
    "        if verb:\n",
    "            self.memory[2] = verb\n",
    "           \n",
    "        return self.run()\n",
    "            \n",
    "            \n",
    "    def run(self):\n",
    "        instr = self.memory[self.pc]\n",
    "        self.stalled = False\n",
    "\n",
    "        while int(instr) != opcode_terminate and not self.stalled:\n",
    "            (op, modes) = self.decode_instr(instr)\n",
    "            self.pc = op(modes)\n",
    "            instr = self.memory[self.pc]\n",
    "\n",
    "        return self.memory[0]\n",
    "\n",
    "    #-HELPERS-----------------------------\n",
    "    def try_pop_mode(self, modes):\n",
    "        if len(modes) == 0:\n",
    "            return 0\n",
    "    \n",
    "        return modes.pop()\n",
    "    \n",
    "    def resize_memory(self, target_addr):\n",
    "        self.memory += ([0] * (1 + target_addr - len(self.memory)))\n",
    "    \n",
    "    #-DECODE-INSTRUCTIONS-----------------\n",
    "    def initialise_opcodes(self):\n",
    "        self.opcodes = {\n",
    "            opcode_add: self.op_add,\n",
    "            opcode_mul: self.op_mul,\n",
    "            opcode_read: self.op_read,\n",
    "            opcode_write: self.op_write,\n",
    "            opcode_jump_true: self.op_jump_true,\n",
    "            opcode_jump_false: self.op_jump_false,\n",
    "            opcode_lt: self.op_lt,\n",
    "            opcode_eq: self.op_eq,\n",
    "            opcode_rebase: self.op_rebase\n",
    "        }    \n",
    "    \n",
    "    def decode_instr(self, instr):\n",
    "        instr = str(instr)\n",
    "        opcode = self.decode_op(instr)\n",
    "        modes = self.decode_modes(instr)\n",
    "\n",
    "        if not (opcode in self.opcodes):\n",
    "            raise Exception(f\"Invalid opcode {opcode}\")\n",
    "\n",
    "        return (self.opcodes[opcode], modes)    \n",
    "    \n",
    "    def decode_op(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return int(instr[-2:])\n",
    "        return int(instr)\n",
    "    \n",
    "    def decode_modes(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return [ int(d) for d in instr[:-2]]\n",
    "        return []\n",
    "    \n",
    "    #-MICRO-OPS---------------------------\n",
    "    def uop_read(self, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if value >= len(self.memory):\n",
    "                self.resize_memory(value)        \n",
    "\n",
    "            return int(self.memory[value])\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + value >= len(self.memory):\n",
    "                self.resize_memory(self.offset + value)     \n",
    "\n",
    "            return int(self.memory[self.offset + value])\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            return int(value)\n",
    "\n",
    "        else:\n",
    "            raise Exception(\"UNKNOWN MODE\")\n",
    "\n",
    "    def uop_write(self, dst, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if dst >= len(self.memory):\n",
    "                self.resize_memory(dst)\n",
    "            self.memory[dst] = value\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + dst >= len(self.memory):\n",
    "                self.resize_memory(self.offset + dst)     \n",
    "\n",
    "            self.memory[self.offset + dst] = value\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            raise Exception(f\"cannot write {value} to literal {dst}\")\n",
    "\n",
    "\n",
    "    def uop_cond_jump(self, modes, cond):\n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest_raw = int(self.memory[self.pc + 2])\n",
    "        dest = self.uop_read(dest_raw, dest_mode)\n",
    "\n",
    "        if cond(param):\n",
    "            return dest\n",
    "\n",
    "        return self.pc + 3\n",
    "\n",
    "    def uop_cmp(self, modes, cmp):\n",
    "\n",
    "        param0_mode = self.try_pop_mode(modes)\n",
    "        param0_raw = int(self.memory[self.pc + 1])\n",
    "        param0 = self.uop_read(param0_raw, param0_mode)\n",
    "\n",
    "        param1_mode = self.try_pop_mode(modes)\n",
    "        param1_raw = int(self.memory[self.pc + 2])\n",
    "        param1 = self.uop_read(param1_raw, param1_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        if cmp(param0, param1):\n",
    "            self.uop_write(dest, 1, dest_mode)\n",
    "        else:\n",
    "            self.uop_write(dest, 0, dest_mode)\n",
    "\n",
    "        return self.pc + 4\n",
    "        \n",
    "    \n",
    "    #-OPCODES-----------------------------\n",
    "    def op_add(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) + int(arg1)), dest_mode)\n",
    "        return self.pc + 4\n",
    "\n",
    "    def op_mul(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) * int(arg1)), dest_mode)    \n",
    "        return self.pc + 4\n",
    "\n",
    "\n",
    "    def op_read(self, modes):\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 1])\n",
    "        \n",
    "        # if the input buffer is empty, we should \"stall\" and \n",
    "        # resume later\n",
    "        if not self.input_buffer:\n",
    "            self.stalled = True\n",
    "            return self.pc\n",
    "        \n",
    "        val = self.input_buffer.pop()\n",
    "\n",
    "        self.uop_write(dest, str(val), dest_mode)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_write(self, modes):\n",
    "        src_mode = self.try_pop_mode(modes)\n",
    "        src_raw = int(self.memory[self.pc + 1])\n",
    "        src = self.uop_read(src_raw, src_mode)\n",
    "\n",
    "        self.output_buffer.append(src)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_jump_true(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x != 0)\n",
    "\n",
    "\n",
    "    def op_jump_false(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x == 0)\n",
    "\n",
    "    def op_lt(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x < y)\n",
    "\n",
    "    def op_eq(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x == y)\n",
    "\n",
    "    def op_rebase(self, modes):    \n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        self.offset += param\n",
    "\n",
    "        return self.pc + 2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "217\n"
     ]
    }
   ],
   "source": [
    "raw_code = open(\"puzzle_input/day19.txt\", \"r\").read()\n",
    "code = raw_code.split(\",\")\n",
    "\n",
    "points_affected = 0\n",
    "output = []\n",
    "output_buffer = []\n",
    "\n",
    "for row in range(0,50):\n",
    "    current_row = []\n",
    "    for col in range(0, 50):\n",
    "        cpu = IntCodeCpu(code)\n",
    "        cpu.start([row, col], output_buffer)\n",
    "        assert 1 == len(output_buffer)\n",
    "        points_affected += output_buffer[0]\n",
    "        output_buffer = []\n",
    "    output.append(current_row)\n",
    "\n",
    "print(points_affected)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "6840937\n"
     ]
    }
   ],
   "source": [
    "def in_beam(row, col):\n",
    "    # is it necessary to create new cpu each time?\n",
    "    cpu = IntCodeCpu(code)\n",
    "    output_buffer = []\n",
    "    input_buffer = [row, col]\n",
    "    cpu.start(input_buffer, output_buffer)\n",
    "    assert 1 == len(output_buffer)\n",
    "    return output_buffer[0] == 1\n",
    "\n",
    "def beam_at_row(row, start_col):\n",
    "    beam_start = start_col\n",
    "    beam_width = 0\n",
    "    col = start_col\n",
    "    \n",
    "    # mine has an empty area in rows\n",
    "    # 1 and 2 so use precomputed values\n",
    "    # yes this is yet another horrible hack\n",
    "    if row < 3:\n",
    "        return [\n",
    "            (0,1),\n",
    "            (0,0),\n",
    "            (0,0)\n",
    "        ][row]\n",
    "    \n",
    "    # seek to the beam start\n",
    "    while not in_beam(row, col):\n",
    "        beam_start += 1\n",
    "        col += 1\n",
    "    \n",
    "    # continue to beam end\n",
    "    while in_beam(row, col):\n",
    "        beam_width += 1\n",
    "        col += 1\n",
    "    \n",
    "    return (beam_start, beam_width)\n",
    "\n",
    "    \n",
    "def beam_can_contain_cube(beam_data, cube_size):\n",
    "    current_beam_index = len(beam_data) - 1\n",
    "    current_beam_start, current_beam_width = beam_data[current_beam_index]\n",
    "    \n",
    "    # so we need to check if:\n",
    "    # 0. there have been at least enough beams (we'll be looking back in the beam_data by cube_size)\n",
    "    if current_beam_index - cube_size < 0:\n",
    "        return False\n",
    "    \n",
    "    # 1. the beam width starting at the current line < cube_size\n",
    "    if current_beam_width < cube_size:\n",
    "        return False\n",
    "    \n",
    "    # 2. the same point at line (current_beam_index - cube_size) is in beam\n",
    "    start_beam_index = current_beam_index - (cube_size - 1)\n",
    "    start_beam_start, start_beam_width = beam_data[start_beam_index]\n",
    "    if current_beam_start >= start_beam_start + start_beam_width:\n",
    "        return False\n",
    "    \n",
    "    # 3. there's enough beam at that ^^^ line afterwards to contain the beam\n",
    "    if current_beam_start + cube_size > start_beam_start + start_beam_width:\n",
    "        return False\n",
    "    \n",
    "    return True\n",
    "    \n",
    "#################\n",
    "target_cube = 100\n",
    "    \n",
    "# seed our beam with a few rows to get going\n",
    "beam = [ beam_at_row(n, 0) for n in range(0, 20) ]    \n",
    "current_beam = len(beam)\n",
    "start_col = 2\n",
    "\n",
    "while not(beam_can_contain_cube(beam, target_cube)):\n",
    "    current = beam_at_row(current_beam, start_col-1)\n",
    "    start_col, current_size = current\n",
    "    beam.append(current)\n",
    "    current_beam += 1\n",
    "    \n",
    "cube_end_row = current_beam - 1\n",
    "cube_start_row = cube_end_row - (target_cube - 1)\n",
    "current_beam_start_col, current_beam_width = beam[-1]\n",
    "print((current_beam_start_col * 10000) + cube_start_row)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
